package com.perforce.hive.artifact;

import com.perforce.hive.api.ChangeModel;
import com.perforce.hive.api.artifact.ArtifactInterface;
import com.perforce.hive.common.ClientFactory;
import com.perforce.hive.common.ClientHelper;
import com.perforce.hive.common.ClientPooledObject;
import com.perforce.hive.connection.ConnectionConfig;
import com.perforce.hive.connection.ConnectionSession;
import com.perforce.hive.depot.DepotModel;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.core.IChangelist;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.FileSpecOpStatus;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.impl.generic.client.ClientView;
import com.perforce.p4java.impl.generic.client.ClientView.ClientViewMapping;
import com.perforce.p4java.impl.generic.core.Changelist;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.option.changelist.SubmitOptions;
import com.perforce.p4java.option.client.ReconcileFilesOptions;
import com.perforce.p4java.option.client.ReopenFilesOptions;
import com.perforce.p4java.option.client.SyncOptions;
import com.perforce.p4java.option.server.OpenedFilesOptions;
import com.perforce.p4java.server.IOptionsServer;
import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.pool2.impl.GenericObjectPool;
import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import spark.Request;
import spark.Response;

import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public abstract class DefaultArtifact implements ArtifactInterface, AutoCloseable {

	private static Logger logger = LoggerFactory.getLogger(DefaultArtifact.class);
	private static String PATH_ID = "Hive Artifact:";

	private static int timeout = ConnectionConfig.get().getPoolTimeout();
	private static GenericObjectPool<ClientPooledObject> pool;

	static {
		GenericObjectPoolConfig config = new GenericObjectPoolConfig();
		int size = ConnectionConfig.get().getProxyPool();
		config.setMaxTotal(size);
		pool = new GenericObjectPool<>(new ClientFactory(), config);
	}

	protected final IOptionsServer p4;

	private IClient client;

	protected String repoID;
	protected Path path;

	// POST multipart form upload
	protected List<FileItem> fileItems = new ArrayList<>();

	// PUT body form upload
	protected String filename;

	/**
	 * Constructor for user based uploads
	 *
	 * @param request
	 * @param response
	 * @throws Exception
	 */
	public DefaultArtifact(Request request, Response response) throws Exception {
		this.p4 = ConnectionSession.get(request, response);

		String method = request.requestMethod().toUpperCase();
		String type = request.headers("Content-Type");
		if (type != null && type.startsWith("multipart")) {
			method = "MULTI";
		}

		switch (method) {
			case "PUT":
				processUrl(request);
				client = privateClient();
				configureClient();
				break;

			case "POST":
				break;

			case "MULTI":
				processForm(request);
				client = privateClient();
				configureClient();
				break;

			default:
				logger.warn("Method not supported: " + method);
		}
	}

	/**
	 * Constructor for Proxy based uploads
	 *
	 * @param p4
	 * @throws Exception
	 */
	public DefaultArtifact(IOptionsServer p4) throws Exception {
		this.p4 = p4;
		client = pool.borrowObject(timeout).getObject();
		logger.info("clientPool borrow: " + client.getName());
	}

	@Override
	public void close() throws Exception {
		if(getClient() == null) {
			return;
		}

		pool.returnObject(new ClientPooledObject(client));
		logger.info("clientPool close: " + client.getName());
		client = null;
		p4.setCurrentClient(null);
	}

	public IClient getClient() {
		return client;
	}

	// Client for the user uploading the file, not a proxy client from the pool.
	private IClient privateClient() throws Exception {
		IClient client = ClientHelper.newClient(p4);

		File root = new File(client.getRoot());
		if (root.exists()) {
			FileUtils.deleteDirectory(root);
		}
		if (!root.exists() && !root.mkdirs()) {
			logger.warn("Failed to create upload directory.");
		}
		return client;
	}

	protected void configureClient() throws Exception {
		// reload client (pool uses a different p4 connection)
		IClient iclient = p4.getClient(client.getName());
		p4.setCurrentClient(iclient);

		// set client view mappings
		Path depotPath = getPath();
		String unixPath = depotPath.toString();
		ClientHelper.mapClientPath(iclient, unixPath);
		client = iclient;
	}

	/**
	 * Implementation processes the content of the multipart form, building a
	 * list of the files to upload and any parameters used to generate the path.
	 *
	 * @param request
	 * @throws FileUploadException
	 */
	protected abstract void processForm(Request request) throws Exception;

	/**
	 * Process a PUT request, e.g:
	 * <p>
	 * curl -v -u admin:Password --upload-file file
	 * http://localhost:4567/view/depot/path/file
	 *
	 * @param request
	 * @throws Exception
	 */
	protected void processUrl(Request request) throws Exception {
		String base = request.pathInfo();
		List<String> parts = new ArrayList<>();
		parts.addAll(Arrays.asList(base.split("/")));
		parts.remove(0);
		parts.remove(0);
		filename = parts.remove(parts.size() - 1);

		String dir = String.join("/", parts);
		// TODO hard code depot name (should be part of the url domain or path)
		path = Paths.get("docker", dir);
	}

	/**
	 * Provide the location of upload directory, used to set Perforce Workspace
	 * root. e.g. 'depot/group/asset/ver'
	 *
	 * @return Workspace root
	 */
	public Path getPath() {
		return path;
	}

	/**
	 * Provide the root location of the index. A Perforce worksapce root with
	 * all the index files.
	 *
	 * @return
	 */
	public abstract DepotModel getDepot() throws P4JavaException;

	/**
	 * Build the index, updating the files in the workspace.
	 *
	 * @throws Exception
	 */
	public abstract void buildIndex() throws Exception;

	@Override
	public void proxyUpload() throws Exception {
		try {
			List<String> upload = populateMultiPartPost();
			int pending = updatePending(upload);
			submitPending(pending, "Submitted by Hive Proxy.");
		} catch (Exception e) {
			throw e;
		} finally {
			// TODO xxx nesting in ProxySite.fetch   close();
		}
	}

	@Override
	public List<String> populateMultiPartPost() throws Exception {

		List<String> list = new ArrayList<>();

		File root = new File(client.getRoot());

		for (FileItem item : fileItems) {
			String fileName = FilenameUtils.getName(item.getName());
			File file = new File(root, fileName);
			item.write(file);
			list.add(file.getPath());
			logger.info("uploading: " + fileName);
		}
		return list;
	}

	@Override
	public List<String> populateMonolithicPut(Request request) throws Exception {

		List<String> list = new ArrayList<>();

		File root = new File(client.getRoot());
		Path path = Paths.get(root.getPath(), filename);

		HttpServletRequest rawRequest = request.raw();
		rawRequest.setCharacterEncoding("ISO-8859-1");
		BufferedReader reader = rawRequest.getReader();

		int n;
		char[] buffer = new char[65536];
		while ((n = reader.read(buffer)) != -1) {
			String string = new String(buffer).substring(0, n);
			byte[] bytes = string.getBytes("ISO-8859-1");
			Files.write(path, bytes, StandardOpenOption.CREATE, StandardOpenOption.APPEND);
		}
		reader.close();

		list.add(path.toString());
		logger.info("uploading: " + path);
		return list;
	}

	@Override
	public int updatePending(List<String> upload) throws Exception {
		// create new pending change and add description
		IChangelist change = new Changelist();
		change.setDescription(PATH_ID + getPath());
		change = client.createChangelist(change);

		// add files to pending change
		String[] array = upload.stream().toArray(String[]::new);
		List<IFileSpec> files = FileSpecBuilder.makeFileSpecList(array);

		// virtual sync -k files
		SyncOptions syncOpts = new SyncOptions();
		syncOpts.setClientBypass(true);
		client.sync(files, syncOpts);

		// reconcile files
		ReconcileFilesOptions statusOpts = new ReconcileFilesOptions();
		statusOpts.setOutsideAdd(true);
		statusOpts.setOutsideEdit(true);
		statusOpts.setChangelistId(change.getId());

		List<IFileSpec> statusSpec = client.reconcileFiles(files, statusOpts);
		validateFileSpecs(statusSpec, "no file(s) to reconcile.");

		return change.getId();
	}

	@Override
	public ChangeModel submitPending(int id, String desc) throws Exception {
		// get current workspace (from pending change)
		IChangelist change = p4.getChangelist(id);
		String clientName = change.getClientId();
		IClient iclient = p4.getClient(clientName);
		p4.setCurrentClient(iclient);

		// !HACK! extract path and repoID from pending description
		String path = change.getDescription();
		path = path.replaceAll("(\\r|\\n)", "");
		path = path.substring(PATH_ID.length());
		String[] parts = path.split("/");
		repoID = parts[0];

		change.setDescription(desc);

		// submit pending change
		SubmitOptions submitOpts = new SubmitOptions();
		List<IFileSpec> submitted = change.submit(submitOpts);
		validateFileSpecs(submitted, "Submitted as change", "No files to submit.");
		logger.info("submitted.");

		long c = findSubmittedChange(submitted);
		ChangeModel changeModel = new ChangeModel(path, c);
		return changeModel;
	}

	private long findSubmittedChange(List<IFileSpec> submitted) {
		long change = 0;
		for (IFileSpec spec : submitted) {
			if (spec.getOpStatus() != FileSpecOpStatus.VALID) {
				String msg = spec.getStatusMessage();
				String cng = "Submitted as change ";
				if (msg.startsWith(cng)) {
					try {
						String id = msg.substring(cng.length());
						change = Long.parseLong(id);
					} catch (NumberFormatException e) {
						change = -1;
					}
				}
			}
		}
		return change;
	}

	protected void versionIndex() throws Exception {
		// fetch depot from implementation
		DepotModel depot = getDepot();

		File root = new File("index", depot.getName());
		String clientName = ConnectionConfig.get().getClientPrefix();
		clientName += depot.getName() + ".index";
		logger.info("index with client: " + clientName);

		// create/setup workspace
		IClient iclient = p4.getClient(clientName);
		Client implClient = new Client(p4);
		implClient.setName(clientName);
		p4.createClient(implClient);
		iclient = p4.getClient(clientName);
		iclient.setRoot(root.getAbsolutePath());
		iclient.setOwnerName(p4.getUserName());

		// create LHS 'depot' mapping
		String lhs = depot.getPath() + "/...";

		// create RHS 'client' mapping
		String rhs = "//" + clientName + "/...";

		// add view mapping
		ClientView clientView = new ClientView();
		String line = lhs + " " + rhs;
		ClientViewMapping entry = new ClientViewMapping(0, line);
		clientView.addEntry(entry);
		iclient.setClientView(clientView);
		iclient.update();

		// select workspace
		p4.setCurrentClient(iclient);

		// build file revision spec
		String ws = "//" + iclient.getName() + "/...";
		List<IFileSpec> files = FileSpecBuilder.makeFileSpecList(ws);

		// sync files
		SyncOptions syncOpts = new SyncOptions();
		List<IFileSpec> syncSpec = iclient.sync(files, syncOpts);
		validateFileSpecs(syncSpec, "file(s) up-to-date");

		// call implementation index here to update files
		buildIndex();

		// add files to default pending change
		ReconcileFilesOptions statusOpts = new ReconcileFilesOptions();
		statusOpts.setOutsideAdd(true);
		statusOpts.setOutsideEdit(true);

		List<IFileSpec> statusSpec = iclient.reconcileFiles(files, statusOpts);
		validateFileSpecs(statusSpec, "no file(s) to reconcile");

		if (isOpened(iclient)) {
			// create new pending change and add description
			IChangelist change = new Changelist();
			change.setDescription("Indexing:" + getPath());
			change = iclient.createChangelist(change);

			// move files from default change
			ReopenFilesOptions reopenOpts = new ReopenFilesOptions();
			reopenOpts.setChangelistId(change.getId());
			iclient.reopenFiles(files, reopenOpts);

			// submit pending change
			SubmitOptions submitOpts = new SubmitOptions();
			List<IFileSpec> submitted = change.submit(submitOpts);
			validateFileSpecs(submitted, "Submitted as change");
			logger.info("indexed updated.");
		} else {
			logger.info("indexed unchanged.");
		}
	}

	private boolean isOpened(IClient iclient) throws Exception {
		String ws = "//" + iclient.getName() + "/...";
		List<IFileSpec> files = FileSpecBuilder.makeFileSpecList(ws);

		OpenedFilesOptions openOps = new OpenedFilesOptions();
		List<IFileSpec> open = iclient.openedFiles(files, openOps);
		for (IFileSpec file : open) {
			if (file != null && file.getAction() != null) {
				return true;
			}
		}
		return false;
	}

	@Override
	public void disconnect() throws Exception {
		p4.disconnect();
	}

	/**
	 * Look for a message in the returned FileSpec from operation.
	 *
	 * @param fileSpecs
	 * @param ignore
	 * @return
	 * @throws Exception
	 */
	private static void validateFileSpecs(List<IFileSpec> fileSpecs, String... ignore) throws Exception {

		boolean abort = false;

		ArrayList<String> ignoreList = new ArrayList<String>();
		ignoreList.addAll(Arrays.asList(ignore));

		for (IFileSpec fileSpec : fileSpecs) {
			FileSpecOpStatus status = fileSpec.getOpStatus();
			if (status != FileSpecOpStatus.VALID) {
				String msg = fileSpec.getStatusMessage();

				// superfluous p4java message
				boolean unknownMsg = true;
				for (String istring : ignoreList) {
					if (msg.contains(istring)) {
						// its a known message
						unknownMsg = false;
						break;
					}
				}

				// check and report unknown message
				if (unknownMsg) {
					logger.warn("P4JAVA: " + msg);
					if (status == FileSpecOpStatus.ERROR || status == FileSpecOpStatus.CLIENT_ERROR) {
						abort = true;
					}
				}
			}
		}

		if (abort) {
			String msg = "P4JAVA: Error(s)";
			throw new P4JavaException(msg);
		}
	}
}
